
// D/Objective-C Cocoa Bindings
// Copyright © 2011  Michel Fortin
// <http://michelf.com/d-objc/>
//
// This file is published under the terms of the Boost Software License 1.0.
//
// Boost Software License - Version 1.0 - August 17th, 2003
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
//
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

module appkit.application;

import appkit.window;
import appkit.menu;
import appkit.responder;
import appkit.image;

import foundation.object;
import foundation.string;


/**
 * NSApplication is the central piece managing events, run loops, and 
 */
extern (Objective-C)
class NSApplication : NSResponder {

	/**
	 * Create the application object. You can create only one application
	 * per process, and the application may not be destroyed.
	 */
	protected this();
	
	/**
	 * Return the shared application instance. This will create a new 
	 * application if it does not yet exist.
	 */
	static NSApplication sharedApplication() @property [sharedApplication];
	
	/**
	 * Run the application event loop.
	 */
	void run() [run];
	
	/**
	 * This method is called before starting the event loop.
	 */
	void finishLaunching() [finishLaunching];
	
	/**
	 * Activate the appliation.
	 */
	void activate(bool ignoringOtherApps = false) [activateIgnoringOtherApps:];
	
	
	// Actions

	void terminate(ObjcObject sender) [terminate:];
	
	void hideOtherApplications(Object sender) [hideOtherApplications:];
	
	
	// Special Menus

	@property {
		void mainMenu(NSMenu menu) [setMainMenu:];
		NSMenu mainMenu() [mainMenu];
	}
	
	
	// Note: -[NSApplication appleMenu] and -[NSApplication setAppleMenu:] 
	// were removed from headers in Tiger for some reason, but are still 
	// needed to set programatically the application menu, so they are here.

	@property {
		void appleMenu(NSMenu menu) [setAppleMenu:];
		NSMenu appleMenu() [appleMenu];
	
		void windowsMenu(NSMenu menu) [setWindowsMenu:];
		NSMenu windowsMenu() [windowsMenu];

		void servicesMenu(NSMenu menu) [setServicesMenu:];
		NSMenu servicesMenu() [servicesMenu];
	}

	// Other propreties.
	
	@property {
		void applicationIconImage(NSImage image) [setApplicationIconImage:];
		NSImage applicationIconImage() [applicationIconImage];
	}
	
	
	// Application delegate.
	
	// deleg proprety (in replacement for "delegate" which is a keyword in D).

	@property {
		void deleg(ObjcObject object) [setDelegate:];
		ObjcObject deleg() [delegate];
	}


	// NSBundle AppKit Additions
	/*
	static bool loadNib(NSString name, Object owner) {
		return NSBundle.invokeObjcSuperClass!(bool, "loadNibNamed:owner:")(name, owner);
	}
	
	static bool loadNib(char[] name, Object owner) {
		return loadNib(name.toNSString(), owner);
	}
	*/
	
	enum TerminateReply { CANCEL = 0, NOW = 1, LATER = 2 }

}


extern (C) int NSApplicationMain(int argc, const(char*)* argv);

int NSApplicationMain(string[] args) {
	auto argp = new immutable(char)*[args.length];
	foreach (i, arg; args)
		argp[i] = arg.ptr;
	
	return NSApplicationMain(argp.length, argp.ptr);
}
