
// D/Objective-C Cocoa Bindings
// Copyright © 2011  Michel Fortin
// <http://michelf.com/d-objc/>
//
// This file is published under the terms of the Boost Software License 1.0.
//
// Boost Software License - Version 1.0 - August 17th, 2003
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
//
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

module foundation.range;

import foundation.string;


extern (C) {
nothrow:

	NSRange NSUnionRange(NSRange range1, NSRange range2);
	NSRange NSIntersectionRange(NSRange range1, NSRange range2);
	NSString NSStringFromRange(NSRange range);
	NSRange NSRangeFromString(NSString aString);
	
}


struct NSRange {
nothrow:

    uint location;
    uint length;
	
	// Commodity function to work with d slices.
	alias location start;
	uint end() { return location + length; }
	void end(uint value) { length = end - location; }
	
	bool locationInRange(uint loc) { return loc - location < length; }
	
	static alias NSUnionRange rangeFromUnion;
	static alias NSIntersectionRange rangeFromIntersection;
	
	static NSRange opSlice(uint from, uint to)
	in {
		assert(to >= from);
	}
	body {
		NSRange range;
		range.location = from;
		range.length = to - from;
		return range;
	}
	
	NSRange unite(NSRange range) {
		return NSUnionRange(this, range);
	}
	
	NSRange intersect(NSRange range) {
		return NSIntersectionRange(this, range);
	}
		
//	NSString toNSString() { 
//		return NSString.objcWrap(NSStringFromRange(*this));
//	}
//	
//	char[] toString() {
//		return toNSString.toString;
//	}
//	
//	static NSRange fromString(NSString string) {
//		return NSRangeFromString(string.objcInstance);
//	}
//	
//	static NSRange fromString(char[] string) {
//		return NSRange.fromString(.toNSString(string));
//	}

}

unittest {

	NSRange range = NSRange[1..3];
	assert(range.location == 1);
	assert(range.length == 2);
	assert(range.end == 3);

}


alias NSRange* NSRangePointer;

static assert(NSRange.sizeof == 8);



//@interface NSValue (NSValueRangeExtensions)
//
//+ (NSValue *)valueWithRange:(NSRange)range;
//- (NSRange)rangeValue;
//
//@end

