
// D/Objective-C Cocoa Bindings
// Copyright © 2011  Michel Fortin
// <http://michelf.com/d-objc/>
//
// This file is published under the terms of the Boost Software License 1.0.
//
// Boost Software License - Version 1.0 - August 17th, 2003
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
//
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

module foundation.string;

import foundation.runtime;
import foundation.object;
import foundation.range;
//import foundation.dictionary;
//import foundation.array;

//import std.string : cmp, toString, toStringz;
//import std.utf : validate, toUTF8, toUTF16;


/**
 * Wrapper for Objective-C class NSString.
 *
 * Subclassing notes: NSString is a cluster class. If you subclass it, you need
 * to create your own Objective-C instance first with objcInstanceAlloc and pass it as 
 * the argument for the NSString constructor. Otherwise, if you call another
 * NSString constructor which creates the Objective-C instance for you, that
 * instance will not be a capsule for your subclass.
 */
extern (Objective-C)
pragma (objc_takestringliteral)
class NSString : NSObject {

	/** Encoding values supported by NSString. */
	enum Encoding : NSUInteger {
		ASCII = 1,		/* 0..127 only */
		NEXTSTEP = 2,
		JAPANESE_EUC = 3,
		UTF8 = 4,
		ISO_LATIN_1 = 5,
		SYMBOL = 6,
		NON_LOSSY_ASCII = 7,
		SHIFT_JIS = 8,
		ISO_LATIN_2 = 9,
		UNICODE = 10,
		WINDOWS_CP1251 = 11,    /* Cyrillic; same as AdobeStandardCyrillic */
		WINDOWS_CP1252 = 12,    /* WinLatin1 */
		WINDOWS_CP1253 = 13,    /* Greek */
		WINDOWS_CP1254 = 14,    /* Turkish */
		WINDOWS_CP1250 = 15,    /* WinLatin2 */
		ISO_2022_JP = 21,         /* ISO 2022 Japanese encoding for e-mail */
		MAC_OS_ROMAN = 30,

		PROPRIETARY = 65536,    /* Installation-specific encoding */
		
		// Note: NSString also support encodings declared in Core Foundation,
		// perhaps they should be added here.
	}
	
	alias NSUInteger CompareOptions;
	
	
	/**
	 * Get the number of characters in the string. Note that UTF-16 surrogate 
	 * pairs are counted as two, so this is not always reflect the number
	 * of characters in the string.
	 *
	 * Subclasses must override this method.
	 */
	NSUInteger length() @property [length];
	
	/**
	 * Return character found at index.
	 *
	 * Subclasses must override this method.
	 */
	wchar opIndex(NSUInteger index) [characterAtIndex:];


	// NSStringExtensionMethods
	
	/**
	 * Copy characters into buffer. Buffer must be large enough for the length 
	 * of the string.
	 */
	void getCharacters(wchar* buffer) [getCharacters:];
	
	/**
	 * Copy characters from a specified range into buffer. Buffer must be 
	 * large enough for the length range.
	 *
	 * Subclasses should override this method if performance is important.
	 */
	void getCharacters(wchar* buffer, NSRange range) [getCharacters:range:];


	/**
	 * Slice the string, returning a substring of the characters between 
	 * indexes from and to.
	 */
	final NSString opSlice(NSUInteger from, NSUInteger to) {
		return substringWithRange(NSRange[from..to]);
	}
	
	/** Convert arguments for opSlice(NSUInteger, NSUInteger). */
	NSString substringWithRange(NSRange range) [substringWithRange:];
	
	
	/**
	 * Compare character of this string to those of the given string,
	 * until a difference is found or the end of the string is reached.
	 */
	NSComparisonResult compare(NSString string) [compare:];
	
	/**
	 * Compare each character of this string to those of the given string
	 * according to options, until a difference is found or the end of the 
	 * string is reached.
	 */
	NSComparisonResult compare(NSString string, CompareOptions options) [compare:options:];
	
	/**
	 * Compare each character of this string within range to those of the 
	 * given string according to options, until a difference is found or 
	 * the end of the string is reached.
	 */
	NSComparisonResult compare(NSString string, CompareOptions options, NSRange range) [compare:options:range:];
	
	/**
	 * Compare each character of this string within range to those of the 
	 * given string according to options and locale, until a difference is 
	 * found or the end of the string is reached.
	 */
//	NSComparisonResult compare(NSString string, CompareOptions options, NSRange range, NSDictionary locale) [compare:options:range:locale:];


	/**
	 * Compare this with object. If object is of class NSString,
	 * use the compare method to compare string. Otherwise, proceed using 
	 * super.opCmp.
	 */
//	final int opCmp(ObjcObject object) {
//		if (auto string = cast(NSString)object) {
//			return compare(string);
//		}
//		return super.opCmp(object);
//	}

	
	NSComparisonResult caseInsensitiveCompare(NSString string) [caseInsensitiveCompare:];
	
	NSComparisonResult localizedCompare(NSString string) [localizedCompare:];
	
	NSComparisonResult localizedCaseInsensitiveCompare(NSString string) [localizedCaseInsensitiveCompare:];
	
	/**
	 * Check for equality of this string with another NSString.
	 */
	bool isEqualToString(NSString string) [isEqualToString:];

	/**
	 * Tell if this string starts with prefix.
	 */
	bool hasPrefix(NSString prefix) [hasPrefix:];

	/**
	 * Tell if this string ends with sufix.
	 */
	bool hasSuffix(NSString sufix) [hasSuffix:];
	

	/**
	 * Search for a matching substring within this and return the range of the 
	 * first occurence. If substring is not found, the returned range will 
	 * have a length of zero.
	 */
	NSRange rangeOfString(NSString substring) [rangeOfString:];
	
	/**
	 * Search for a matching substring using options within this and return 
	 * the range of the first occurence. If substring is not found, the 
	 * returned range will have a length of zero.
	 */
	NSRange rangeOfString(NSString substring, CompareOptions options) [rangeOfString:options:];
	
	/**
	 * Search for a matching substring using options within a specific range 
	 * of this and return the range of the first occurence. If substring is 
	 * not found, the returned range will have a length of zero.
	 */
	NSRange rangeOfString(NSString substring, CompareOptions options, NSRange range) [rangeOfString:options:range:];


	/**
	 * Create a string from the concatenation of this and string.
	 */
	NSString opCat(NSString string) [stringByAppendingString:];


	/** Give the value of this string when interpreted as a double. */
	double doubleValue() @property [doubleValue];

	/** Give the value of this string when interpreted as a float. */
	float floatValue() @property [floatValue];

	/** Give the value of this string when interpreted as an int. */
	int intValue() @property [intValue];
	

	/**
	 * Give a modified version of this string where lowercase characters are
	 * converted to uppercase.
	 */
	NSString uppercaseString() @property [uppercaseString];

	/**
	 * Give a modified version of this string where uppercase characters are
	 * converted to uppercase.
	 */
	NSString lowercaseString() @property [lowercaseString];

	/**
	 * Give a modified version of this string where characters at the start of 
	 * words are converted to uppercase and all other characters are converted 
	 * to lowercase.
	 */
	NSString capitalizedString() @property [capitalizedString];
		
	
	/**
	 * Get a UTF-8 representation of this string.
	 */
	char* UTF8String() @property [UTF8String];


	// Creation methods

	/** Create an empty string. */
	this() [init];
	
	/** Create a string from bytes using encoding. */
	this(byte* bytes, NSUInteger length, Encoding encoding) [initWithBytes:length:encoding:];

/*
	version (MAC_OS_X_VERSION_10_4) {

		bool writeToURL(NSURL url, bool atomically, NSStringEncoding encoding, NSError error) {
			return invokeObjcSuper!(bool, "writeToURL:atomically:encoding:error:")(url, atomically, encoding, error);
		}
		mixin ObjcBindMethod!(writeToURL, bool, "writeToURL:atomically:encoding:error:", NSURL, bool, NSStringEncoding, NSError);

		bool writeToFile(NSString path, bool atomically, NSStringEncoding encoding, NSError error) {
			return invokeObjcSuper!(bool, "writeToFile:atomically:encoding:error:")(path, atomically, encoding, error);
		}
		mixin ObjcBindMethod!(writeToFile, bool, "writeToFile:atomically:encoding:error:", NSString, bool, NSStringEncoding, NSError);

	}
*/
/*
	static NSString stringWithFormat(A...)(NSString format, A args) {
		return invokeObjcSuperClass!(NSString, "stringWithFormat:", false)(format, args);
	}
	
	static NSString stringWithFormat(A...)(char[] format, A args) {
		return stringWithFormat(format.toNSString(), args);
	}
*/
}

/+
/**
 * Wrapper for Objective-C class NSMutableString.
 *
 * Subclassing notes: NSMutableString is a cluster class. If you subclass it, 
 * you need to create your own Objective-C instance first with objcInstanceAlloc and 
 * pass it as the argument for the NSMutableString constructor. Otherwise, if 
 * you call another NSMutableString constructor which creates the Objective-C 
 * instance for you, that instance will not be a capsule for your subclass.
 */
class NSMutableString : NSString {
	mixin ObjcClusterWrap;
	this(id instance) { super(instance); }
	
	/**
	 * Replace characters between index from and to with characters from string
	 * replacement.
	 */
	void opSliceAssign(NSString replacement, NSUInteger from, NSUInteger to) {
		invokeObjcSuper!(void, "replaceCharactersInRange:withString:")(NSRange[from..to], replacement);
	}
	
	/** ditto */
	void opSliceAssign(char[] replacement, NSUInteger from, NSUInteger to) {
		opSliceAssign(replacement.toNSString(), from, to);
	}
	
	/** Convert arguments for opSliceAssign(NSString, NSUInteger, NSUInteger). */
	private void replaceCharactersInRange(NSRange range, NSString replacement) {
		return opSliceAssign(replacement, range.location, range.end);
	}
	
	mixin ObjcBindMethod!(replaceCharactersInRange, void, "replaceCharactersInRange:withString:", NSRange, NSString);

	
	// NSMutableStringExtensionMethods
	
	/**
	 * Insert string at index.
	 */
	void insert(NSString string, NSUInteger index) {
		invokeObjcSuper!(void, "insertString:atIndex:")(string, index);
	}
	
	/**
	 * Delete character in range.
	 */
	void remove(NSRange range) {
		invokeObjcSuper!(void, "deleteCharacters:")(range);
	}
	
	/** ditto */
	void remove(NSUInteger from, NSUInteger to) {
		remove(NSRange[from..to]);
	}
	
	/** Concatenate string at the end of this one. */
	void opCatAssign(NSString string) {
		invokeObjcSuper!(void, "appendString:")(string);
	}
	
	/** ditto */
	void opCatAssign(char[] string) {
		opCatAssign(string.toNSString());
	}
	
	/** Replace this string content's with string. */
	void opAssign(NSString string) {
		invokeObjcSuper!(void, "setString:")(string);
	}
	
	/** ditto */
	final void opAssign(char[] string) {
		opAssign(string.toNSString());
	}
	
	mixin ObjcBindMethod!(insert, void, "insertString:atIndex:", NSString, NSUInteger);
	mixin ObjcBindMethod!(remove, void, "deleteCharacters:", NSRange);
	mixin ObjcBindMethod!(opCatAssign, void, "appendString:", NSString);
	mixin ObjcBindMethod!(opAssign, void, "setString:", NSString);
		

	// NSMutableString Creation
	
	/**
	 * Create a new mutable string with a given capacity. This may not allocate
	 * memory immediatly, however when memory gets allocated it'll be enough
	 * for the specified capacity.
	 */
	this(NSUInteger capacity) {
		super(objcInstanceAlloc);
		invokeObjcInit!("initWithCapacity:")(capacity);
	}


	// NSString Creation methods

	/** Create an empty string. */
	this() {
		super();
	}

	/** Create a string from UTF-8 characters. */
	this(char[] characters) {
		super(characters);
	}
	
	/** Create a string from bytes using encoding. */
	this(byte[] bytes, Encoding encoding) {
		super(bytes, encoding);
	}

}


unittest {
	scope auto a = "".toNSString();
}


/**
 * Options flags to use with string comparaison, search, and find methods.
 */
enum {
	/** Ignore case difference. */
    NSCaseInsensitiveSearch = 1,
	
	/** Use a byte-for-byte comparator, ignoring character equivalences. */
    NSLiteralSearch = 2,
	
	/** Start searching at the end of the string. */
    NSBackwardsSearch = 4,
	
	/** Search only at either the start or the of the string, depending on NSBackwardSearch. */
    NSAnchoredSearch = 8,
	
	/** Compare numbers using numeric values when multiple contigus digits are found in the string. */
    NSNumericSearch = 64
}


private class D_String(C) : NSString {
	mixin ObjcClusterImpl;

	static assert(is(C == char) || is(C == wchar) || is(C == dchar));	

	/** Character storage array. */
	final C[] storage;
	
	/** Create an empty string. */
	this() {
		super();
	}
	
	/** Create a string with given characters. */
	this(C[] characters)
	in {
		validate(characters);
	}
	body {
		super();
		storage = characters.dup;
	}
	
	
	override wchar opIndex(NSUInteger index) {
		return storage[index];
	}
	
	override NSUInteger length() {
		static if (is(C == wchar)) {
			return storage.length;
		} else {
			// TODO: Optimize to avoid creating a new string.
			return storage.toUTF16().length;
		}
	}


	// (NSStringExtensionMethods)
	
	override void getCharacters(wchar* buffer) {
		buffer[0..storage.length] = storage.toUTF16();
	}
		
	override void getCharacters(wchar* buffer, NSRange range) {
		// TODO: Optimize that: no need to convert the whole string to UTF-16.
		buffer[0..range.length] = storage.toUTF16()[range.location..range.end];
	}
	
	static if (is(C == wchar)) {
	
		override NSString opSlice(NSUInteger from, NSUInteger to) {
			return new D_String(storage[from..to]);
		}
	
	}
	
	alias NSString.compare compare;
	
	static if (is(C == char)) {
		// std.string.cmp only works on char
	
		override NSComparisonResult compare(NSString string) {
			if (auto d_string = cast(D_String)string) {
				auto result = cmp(this.storage, d_string.storage);
				return cast(NSComparisonResult)result;
			}
			return super.compare(string);
		}
		
		// Buggy: range is expressed in term of UTF-16 length, not UTF-8 as this function implies.
//		override NSComparisonResult compare(NSString string, CompareOptions options, NSRange compareRange) {
//			if (options == NSLiteralSearch) {
//				if (auto d_string = cast(D_String)string) {
//					with (compareRange) {
//						// Slice...
//						char[] first  = this.storage[start..end];
//						char[] second = d_string.storage[start..end];
//						// ... and compare
//						auto result = std.string.cmp(first, second);
//						return cast(NSComparisonResult)result;
//					}
//				}
//			}
//			return super.compare(string, options, compareRange);
//		}
	
	}
	
	override bool isEqualToString(NSString string) {
		if (auto d_string = cast(D_String)string) {
			return this.storage == d_string.storage;
		} else {
			return super.isEqualToString(string);
		}
	}

	static if (is(C == char)) {
	
		override NSString opCat(NSString string) {
			return new D_String(storage ~ string.toString);
		}
		
		alias NSString.opCat opCat;
	
	} else {
	
		override NSString opCat(NSString string) {
			if (auto d_string = cast(D_String)string) {
				return opCat(d_string);
			}
			return super.opCat(string);
		}
		
		alias NSString.opCat opCat;
	
	}
	
	override char[] toString() {
		return storage.toUTF8();
	}
	
	static if (is(C == char)) {
		
		override int opCmp(C[] string) {
			return cmp(this.storage, string);
		}

		D_String opCat(C[] string) {
			return new D_String(this.storage ~ string);
		}
	
	}
	
}


import cocoa.foundation.autoreleasepool;

unittest {
	scope NSAutoreleasePool pool = new NSAutoreleasePool;

	scope a = new NSString("string");
	scope b = new D_String!(char)("string");
	scope c = "string".toNSString();
	
	assert(a.isEqualToString(b));
	assert(a == b);
	assert(b == c);
	
	assert(a ~ "test" == "stringtest");
	
	// Check operations with multibyte UTF-8 sequences.
	assert("Français".toNSString().length == 8);
	assert("русский язык".toNSString().length == 12);
	assert((a ~ "日本語").length == a.length + 3);

}
+/

